import intel.pcsdk.PXCUPipeline;

import java.awt.Color;
import java.util.ArrayList;

import processing.core.PApplet;
import processing.core.PGraphics;
import processing.core.PImage;
import processing.core.PShape;
import processing.core.PVector;

  ////////////////////////////////////////
  boolean intelEnabled = true;
  ////////////////////////////////////////

  IntelCameraCalss cintel;

  boolean debugViewOn = false;

  PVector ssize = new PVector(1680, 1050);
  int swidth, sheight;

  // PARALLAX

  // Tiles
  int numTiles = 5;
  PVector tileSize = new PVector(400, ssize.y);
  float totalTileWidth = tileSize.x * numTiles;
  float totalTileHeight = tileSize.y;
  float tilesXrange = totalTileWidth / 2 - tileSize.x / 2;
  float tilesX = 0;

  // Screen change progressbar
  float progressBarVelocity = 0.03f * 1.25f * 1.8f;
  float progressBarDevelocity = -progressBarVelocity * 0.5f;
  float progressBarValue = 0;
  boolean progressBarFromTop = false;

  // Views
  static int COLLECTIONSVIEW = 1;
  static int DETAILSVIEW = 2;
  int currentView = 2;

  // Screen Transform
  SkewedImage skewedImage;

  // Program Dummy Values
  int _indexOfCentreItem = 0;
  PImage imgScreen;

  ArrayList<PShape> objs;

  public void setup() {
    swidth = (int) ssize.x;
    sheight = (int) ssize.y;
    size(swidth, sheight, OPENGL);
    colorMode(HSB, 100);

    this.frame.setBackground(Color.BLACK);
    
    if (intelEnabled) {
      cintel = new IntelCameraCalss(this);
    }

    skewedImage = new SkewedImage();
    skewedImage.drawDebug = false;

    objs = new ArrayList<PShape>();
    PShape _obj;
    _obj = loadShape("horse-obj.obj");
    objs.add(_obj);
    _obj = loadShape("sea urchine1.obj");
    objs.add(_obj);
    _obj = loadShape("Mushroom.obj");
    objs.add(_obj);
    _obj = loadShape("rocket.obj");
    objs.add(_obj);
    _obj = loadShape("PlantObject.obj");
    objs.add(_obj);
  }

  public void draw() {
    //    skewedImage.drawDebug = debugViewOn;
    if (!debugViewOn) {
      noCursor();
    } else {
      cursor();
    }

    background(0);

    float surfaceAngle = 0;
    float surfaceAngleY = 0;

    // Update IntelCamera
    if (intelEnabled) {
      cintel.update();
      surfaceAngle = degrees(cintel.getAngleOfSurface());
      surfaceAngleY = -degrees(cintel.getAngleOfSurfaceY());
    }

    // Update Skewed Image.
    float eyexRange = 500;
    if (intelEnabled) {
      eyexRange = 400;
    }
    float eyeX2 = map(surfaceAngle, 45, -45, -eyexRange + swidth / 2, eyexRange + swidth / 2);
    skewedImage.setEyeX(eyeX2);
    skewedImage.update();

    ////////////////////////////////////
    // Parallax
    background(0);
    int angleRange = 45;
    float inputX = map(surfaceAngle, -angleRange, angleRange, 0, swidth);
    int angleRangeY = 45;
    float inputY = map(surfaceAngleY, angleRangeY, -angleRangeY, 0, sheight);

    if (!intelEnabled) {
      skewedImage.img = updateAndGetParallaxImage(mouseX, mouseY);
    } else {
      skewedImage.img = updateAndGetParallaxImage(inputX, inputY);
    }
    background(0);

    // Draw Skewed Image.
    skewedImage.draw();

  }

  public void keyPressed() {
    if (key == 'd') {
      debugViewOn = !debugViewOn;
    }
  }

  PImage updateAndGetParallaxImage(float inputX, float inputY) {
    // should be a pgraphic.

    ///////////////////////////////////////////////////////////////////////////////////////////////
    // UPDATE
    PVector centralMouse = new PVector(inputX - swidth / 2, inputY - sheight / 2);
    PVector thresholdedMouse = centralMouse.get();

    // Set 'safe' area for centralMouse - as thresholdedMouse
    float thresholdX = swidth * 0.035f;
    if (thresholdedMouse.x > thresholdX) {
      thresholdedMouse.x -= thresholdX;
    } else if (thresholdedMouse.x < -thresholdX) {
      thresholdedMouse.x += thresholdX;
    } else {
      thresholdedMouse.x = 0;
    }
    float thresholdY = sheight * 0.10f;
    if (thresholdedMouse.y > thresholdY) {
      thresholdedMouse.y -= thresholdY;
    } else if (thresholdedMouse.y < -thresholdY) {
      thresholdedMouse.y += thresholdY;
    } else {
      thresholdedMouse.y = 0;
    }
    // Extra condition - don't move y if x is moving
    //    if (thresholdedMouse.x != 0) {
    //      thresholdedMouse.y = 0;
    //    }

    // progressbar
    progressBarValue += progressBarDevelocity;
    progressBarValue = Math.max(0, progressBarValue);
    progressBarValue = Math.min(progressBarValue, 1);

    if (currentView == COLLECTIONSVIEW) {
      // Only in collectionsView
      // update tilesX with velocity from mouse position.
      float inputForTilesX = -thresholdedMouse.x * 0.25f;
      tilesX += inputForTilesX;
      tilesX = Math.max(tilesX, -tilesXrange);
      tilesX = Math.min(tilesX, tilesXrange);
      //tilesX = map(mouseX, 0, swidth, -range, range);
    }

    ///////////////////////////////////////////////////////////////////////////////////////////////
    // DRAW
    background(0, 100);

    // DRAW TILE - DETAILSVIEW
    if (currentView == DETAILSVIEW) {
      int i = _indexOfCentreItem;
      float hue = i * 1f / numTiles * 100;
      strokeWeight(2);
      stroke(hue, 80, 95);
      fill(hue, 20, 30);

      PShape _obj = objs.get(i);

      pushMatrix();
      //      translate(swidth / 2, sheight / 2);
      //      translate(-tileSize.x / 2, -tileSize.y / 2);
      translate(0, 0, -_obj.getDepth() * 2.5f);
      //      rect(0, 0, tileSize.x, tileSize.y);
      rect(0, 0, swidth, sheight);
      popMatrix();

      pushMatrix();
      pushStyle();
      colorMode(RGB, 255);
      lights();
      float h = 66;
      directionalLight(h, h, h, 0, 0, -1);
      ambientLight(102, 102, 102);
      translate(swidth / 2, sheight / 2);
      if ((i == 0) || (i == 4)) {
        // ground based models
        translate(0, sheight * 0.25f);
      }

      //      println(_obj.getWidth()+","+_obj.getHeight());
      float maxDimensionSize = Math.max(_obj.getWidth(), _obj.getHeight());
      float oscale = swidth * 0.5f * 1f / maxDimensionSize * 0.5f;
      scale(oscale);
      rotateZ(PI);
      rotateY(map(inputX, 0, swidth, -PI * 2, PI * 2));

      noFill();

      shape(_obj);
      popStyle();
      popMatrix();

    }

    // DRAW TILES //only in collections view
    if (currentView == COLLECTIONSVIEW) {
      pushMatrix();
      // Translate Centre
      translate(swidth / 2, sheight / 2);
      // Translate to centre of tileView
      translate(-totalTileWidth / 2, -totalTileHeight / 2);
      translate(tilesX, 0);

      float tilesXPctg = map(tilesX, totalTileWidth / 2, -totalTileWidth / 2, 0, 1);
      int curCentralTile = (int) (tilesXPctg * numTiles);
      _indexOfCentreItem = curCentralTile;

      for (int i = 0; i < numTiles; i++) {
        float hue = i * 1f / numTiles * 100;
        strokeWeight(2);
        stroke(hue, 80, 98);
        fill(hue, 80, 75);
        pushMatrix();

        if (i == _indexOfCentreItem) {
          strokeWeight(9);
        }

        translate(i * tileSize.x, 0);
        rect(0, 0, tileSize.x, tileSize.y);
        popMatrix();
      }
      popMatrix();
    }

    // DRAW BUTTONS //changes based on view
    if (currentView == DETAILSVIEW) {
      pushMatrix();
      // Parrallax transform
      float uitranslatex = -thresholdedMouse.x * 0.070f;
      float uitranslatey = -thresholdedMouse.y * 0.300f;
      translate(uitranslatex, uitranslatey);

      // Left, Right Icons
      strokeWeight(6);
      stroke(0, 0, 60);
      fill(0, 0, 90, 90);
      float r = 50; //actually diametre, not radius
      int z = 20;
      float scaleHighlight = 1.2f;
      // Left Icon
      pushMatrix();
      translate(r / 2, sheight / 2, z);
      if (uitranslatex > 0) {
        scale(scaleHighlight);
      }
      ellipse(0, 0, r, r);
      popMatrix();
      // Right Icon
      pushMatrix();
      translate(swidth - r / 2, sheight / 2, z);
      if (uitranslatex < 0) {
        scale(scaleHighlight);
      }
      ellipse(0, 0, r, r);
      popMatrix();

      // Bottom Icon
      pushMatrix();
      translate(swidth / 2, r * 0.25f, z);
      if (uitranslatey > 0) {
        scale(scaleHighlight);
        if (!progressBarFromTop)
          progressBarValue = 0;
        progressBarValue += progressBarVelocity;
        progressBarFromTop = true;
        // progress to next view
        if (progressBarValue >= 1) {
          currentView = COLLECTIONSVIEW;
          progressBarValue = 0;
        }
      }
      rect(-r / 2, -r / 2, r, r);
      popMatrix();

      popMatrix();
    }

    if (currentView == COLLECTIONSVIEW) {
      pushMatrix();
      // Parrallax transform
      float uitranslatex = -thresholdedMouse.x * 0.070f;
      float uitranslatey = -thresholdedMouse.y * 0.300f;
      translate(uitranslatex, uitranslatey);

      // Left, Right Icons
      strokeWeight(6);
      stroke(0, 0, 60);
      fill(0, 0, 90, 90);
      float r = 50; //actually diametre, not radius
      int z = 20;
      float scaleHighlight = 1.2f;
      // Left Icon
      pushMatrix();
      translate(r / 2, sheight / 2, z);
      if (uitranslatex > 0) {
        scale(scaleHighlight);
      }
      ellipse(0, 0, r, r);
      popMatrix();
      // Right Icon
      pushMatrix();
      translate(swidth - r / 2, sheight / 2, z);
      if (uitranslatex < 0) {
        scale(scaleHighlight);
      }
      ellipse(0, 0, r, r);
      popMatrix();

      // Top, Down Icons
      // Top Icon
      //      pushMatrix();
      //      translate(swidth / 2, r * 0.25f, z);
      //      if (uitranslatey > 0) {
      //        scale(scaleHighlight);
      //        if (!progressBarFromTop)
      //          progressBarValue = 0;
      //        progressBarValue += progressBarVelocity;
      //        progressBarFromTop = true;
      //      }
      //      rect(-r / 2, -r / 2, r, r);
      //      popMatrix();
      // Bottom Icon
      pushMatrix();
      translate(swidth / 2, sheight - r * 0.25f, z);
      if (uitranslatey < 0) {
        scale(scaleHighlight);
        if (progressBarFromTop)
          progressBarValue = 0;
        progressBarValue += progressBarVelocity;
        progressBarFromTop = false;
        // progress to next view
        if (progressBarValue >= 1) {
          currentView = DETAILSVIEW;
          progressBarFromTop = true;
          progressBarValue = 0;
        }
      }
      rect(-r / 2, -r / 2, r, r);
      popMatrix();
      popMatrix();
    }

    // DRAW PROGRESS-BAR 
    pushMatrix();
    noStroke();
    strokeWeight(1);
    stroke(0, 0, 100);
    fill(0, 0, 20, 30);
    float h = sheight * progressBarValue;
    if (progressBarValue != 0) {
      if (progressBarFromTop) {
        rect(0, 0, swidth, h);
      } else {
        translate(0, sheight - h);
        rect(0, 0, swidth, h);
      }
    }
    popMatrix();

    // RETURN WHOLE IMAGE
    return this.get();

  }

  /**
   * Custom Intel Camera class for program.
   */
  public class IntelCameraCalss {
    //IntelCamera session
    PXCUPipeline session;
    int[] depthMapSize = new int[2];
    short[] depthMap;
    PImage depthImage;
    int[] uvMapSize = new int[2];
    float[] uvMap;
    int[] rgbMapSize = new int[2];
    PImage rgbImage;
    ArrayList<PVector> pointCloud = new ArrayList<PVector>(); //unnecessary
    boolean drawRGB = false;

    boolean generatingPointCloudAndDepthImage = true;

    // Points to determine surface angle
    float angleOfSurface;
    float angleOfSurfaceY;
    float samplePointDistance = 0.25f; //as a percentage of source width
    int cx, cy, lx, ly, rx, ry, ux, uy, dx, dy; // co-ordinates of central point, and left and right sampling points

    IntelCameraCalss(PApplet papplet) {
      // Initialize pipeline.
      session = new PXCUPipeline(papplet);
      if (!session.Init(PXCUPipeline.DEPTH_QVGA | PXCUPipeline.COLOR_VGA))
        exit();
      if (session.QueryDepthMapSize(depthMapSize)) {
        depthMap = new short[depthMapSize[0] * depthMapSize[1]];
        depthImage = createImage(depthMapSize[0], depthMapSize[1], ALPHA);
      }
      if (session.QueryUVMapSize(uvMapSize)) {
        uvMap = new float[uvMapSize[0] * uvMapSize[1] * 2];
      }
      if (session.QueryRGBSize(rgbMapSize)) {
        rgbImage = createImage(rgbMapSize[0], rgbMapSize[1], RGB);
      }

      cx = depthMapSize[0] / 2;
      cy = depthMapSize[1] / 2;
      int lrdist = (int) (depthMapSize[0] * samplePointDistance);
      lx = cx - lrdist / 2;
      ly = cy;
      rx = cx + lrdist / 2;
      ry = cy;
      ux = cx;
      uy = cy - lrdist / 2;
      dx = cx;
      dy = cy + lrdist / 2;
    }

    void update() {
      if (session.AcquireFrame(false)) {
        pointCloud.clear();

        session.QueryDepthMap(depthMap);
        session.QueryUVMap(uvMap);

        session.QueryRGB(rgbImage);

        rgbImage.loadPixels();

        if (generatingPointCloudAndDepthImage) {
          float remapMouseX = map(mouseY, 0, width, 255, 8192);
          for (int x = 0; x < depthMapSize[0]; ++x) {
            for (int y = 0; y < depthMapSize[1]; ++y) {
              int i = y * depthMapSize[0] + x;

              depthImage.pixels[i] = color(map(depthMap[i], 0, remapMouseX, 0, 255));

              // add point to pointCloud.
              if (depthMap[i] > 0) {
                PVector pointIn3d = getCentred3dPoint(x, y, depthMap[i], depthMapSize[0], depthMapSize[1]);
                pointCloud.add(pointIn3d.get());
                //pointCloud.add( new PVector(x,y,(int)depthMap[i]));
              }
            }
          }
          depthImage.updatePixels();
        }

        session.ReleaseFrame();//VERY IMPORTANT TO RELEASE THE FRAME
      }

      //      int ci = cx + cy * depthMapSize[0];
      //      println(depthMap[ci]);

      // Calculate angle of Surface
      PVector leftPoint = getCentred3dPoint(lx, ly, depthMap[(lx + ly * depthMapSize[0])], depthMapSize[0], depthMapSize[1]);
      PVector rightPoint = getCentred3dPoint(rx, ry, depthMap[(rx + ry * depthMapSize[0])], depthMapSize[0], depthMapSize[1]);
      angleOfSurface = calculateAngleBetweenTwoPoints(leftPoint.x, leftPoint.z, rightPoint.x, rightPoint.z);

      PVector upperPoint = getCentred3dPoint(ux, uy, depthMap[(ux + uy * depthMapSize[0])], depthMapSize[0], depthMapSize[1]);
      PVector downPoint = getCentred3dPoint(dx, dy, depthMap[(dx + dy * depthMapSize[0])], depthMapSize[0], depthMapSize[1]);
      angleOfSurfaceY = calculateAngleBetweenTwoPoints(upperPoint.y, upperPoint.z, downPoint.y, downPoint.z);

    }

    float calculateAngleBetweenTwoPoints(float lx, float ly, float rx, float ry) {
      // Calculate angle
      float dy = ly - ry;
      float dx = lx - rx;
      float angle = (float) Math.atan(dy / dx);
      return angle;
    }

    float getAngleOfSurface() {
      return angleOfSurface;
    }

    float getAngleOfSurfaceY() {
      return angleOfSurfaceY;
    }

    void draw() {
      pushMatrix();
      translate(width / 2, height / 2);
      rotateY(radians(180 + mouseX));

      if (drawRGB) {
        pushMatrix();
        translate(-width / 2, -height / 2);

        // Draw Depth Image
        image(depthImage, 0, 0);

        strokeWeight(1);
        stroke(0, 0, 0);
        fill(255, 255, 255, 150);
        // Draw Dots
        ellipse(cx, cy, 10, 10);
        ellipse(lx, ly, 10, 10);
        ellipse(rx, ry, 10, 10);

        // Draw RGB Image
        pushMatrix();
        translate(320, 0);
        image(rgbImage, 0, 0);
        // Draw Dots
        pushMatrix();
        scale(2f);
        ellipse(cx, cy, 10, 10);
        ellipse(lx, ly, 10, 10);
        ellipse(rx, ry, 10, 10);
        popMatrix();
        popMatrix();

        popMatrix();
      }

      // Draw Point Cloud
      pushMatrix();
      translate(0, 0, 200);
      for (int p = 0; p < pointCloud.size(); ++p) {
        stroke(0, 255, 0);
        PVector pt = (PVector) pointCloud.get(p);
        point(pt.x, pt.y, -pt.z);
      }
      popMatrix();
      popMatrix();
    }

    PVector getCentred3dPoint(int imgx, int imgy, int depthz, int imgWidth, int imgHeight) {

      float tx = imgx - imgWidth / 2;
      float ty = imgy - imgHeight / 2;

      // estimation of realworld distance (in millimetres)
      float realWorldScreenWidthAtGivenDepth = 1.125294165f * depthz - 47.429104743f;

      float txToScreenRatio = tx * 1f / imgWidth;
      float realWorldX = realWorldScreenWidthAtGivenDepth * txToScreenRatio;

      float aspectRatio = imgWidth * 1f / imgHeight;

      float tyToScreenRatio = (ty * 1f / imgHeight) / aspectRatio;
      float realWorldY = realWorldScreenWidthAtGivenDepth * tyToScreenRatio;

      PVector centred3dPoint = new PVector(realWorldX, realWorldY, depthz);

      return centred3dPoint;
    }
  }

  public class SkewedImage {
    PGraphics pg;
    PImage img;
    private float pgScale = 2;
    boolean drawDebug = true;

    float eyeX;

    SkewedImage() {
      pg = createGraphics((int) (swidth * pgScale), (int) (sheight * pgScale), OPENGL);
      eyeX = swidth * 0.5f;
    }

    void setEyeX(float eyeX) {
      this.eyeX = eyeX;
    }

    // drawn so base edge is grounded and sides are perpendicular to ground.
    void update() {
      float rs = 300; //square size

      pg.beginDraw();
      pg.clear();
      pg.smooth();
      pg.scale(pgScale);

      // SETUP CAMERA
      float eyeY = sheight * 0.5f;
      eyeY = sheight / 2 + rs / 2;
      float eyeZ = (float) ((height / 2.0f) / Math.tan(PI * 30.0 / 180.0));

      float centreX = swidth * 0.5f;
      float centreY = sheight * 0.5f;
      float centreZ = 0;

      centreY = eyeY;

      pg.camera(eyeX, eyeY, eyeZ, centreX, centreY, centreZ, 0, 1, 0);

      // perspective change tailored for the P300.
      float eyeXDistFromCentre = Math.abs(eyeX - swidth / 2);
      float aspectPerspectiveRatioMultiplier = 1 + eyeXDistFromCentre * 0.0001f;
      float cameraZ = (float) ((height / 2.0) / Math.tan(PI * 60.0 / 360.0));
      //pg.perspective(PI / 4.70f, width * aspectPerspectiveRatioMultiplier / height, cameraZ / 10.0f, cameraZ * 10.0f); // pg.perspective(PI/3.0f, width*1f/height, cameraZ/10.0f, cameraZ*10.0f);
      pg.perspective(PI / 2.7f, width * aspectPerspectiveRatioMultiplier / height, cameraZ / 10.0f, cameraZ * 10.0f);
      pg.noFill();

      if (drawDebug) {
        pg.pushMatrix();
        pg.translate(swidth / 2, sheight / 2);
        pg.colorMode(HSB, 100);
        pg.strokeWeight(10);
        pg.stroke(20, 90, 100);
        pg.fill(20, 90, 90, 70);
        pg.rect(-rs / 2, -rs / 2, rs, rs);
        pg.popMatrix();
      }
      if (img != null) {
        pg.pushMatrix();
        pg.translate(swidth / 2, sheight / 2 + rs / 2); //~ because eyeY is at sheight/2+rs/2
        pg.scale(0.5f);
        pg.translate(-img.width / 2, -img.height);
        pg.image(img, 0, 0);
        pg.popMatrix();
      }

      pg.endDraw();
    }

    void draw() {
      pushMatrix();
      translate(-(skewedImage.pg.width - swidth) / 2, 0);
      image(skewedImage.pg, 0, 0);
      popMatrix();
    }
  }

